# Copy from https://github.com/mooz/percol/blob/master/percol/tty.py

# -*- coding: utf-8 -*-

import sys
import os


def get_ttyname():
    for f in sys.stdin, sys.stdout, sys.stderr:
        if f.isatty():
            return os.ttyname(f.fileno())
    return None


def reconnect_descriptors(tty):
    target = {}

    stdios = (("stdin", "r"), ("stdout", "w"), ("stderr", "w"))

    tty_desc = tty.fileno()

    for name, mode in stdios:
        f = getattr(sys, name)

        if f.isatty():
            # f is TTY
            target[name] = f
        else:
            # f is other process's output / input or a file

            # save descriptor connected with other process
            std_desc = f.fileno()
            other_desc = os.dup(std_desc)

            # set std descriptor. std_desc become invalid.
            os.dup2(tty_desc, std_desc)

            # set file object connected to other_desc to corresponding one of sys.{stdin, stdout, stderr}
            try:
                target[name] = os.fdopen(other_desc, mode)
                setattr(sys, name, target[name])
            except OSError:
                # maybe mode specification is invalid or /dev/null is specified (?)
                target[name] = None
                print("Error: Failed to open {0}".format(other_desc))

    return target



# coding:utf-8

import json
import os, sys
import signal
import warnings
import curses

# local

try:
    # need when 「python3 gfzs/cmd/demo.py」
    if __name__ == "__main__":
        # https://codechacha.com/ja/how-to-import-python-files/
        sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
        from controller import Controller
        from runtime.config import RuntimeConfig
        from runtime.opts import RuntimeOpts

        if os.environ.get("DEBUG"):
            import debug

    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    else:
        from gfzs.controller import Controller
        from gfzs.runtime.config import RuntimeConfig
        from gfzs.runtime.opts import RuntimeOpts

        if os.environ.get("DEBUG"):
            import gfzs.utils.debug as debug

# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    # https://codechacha.com/ja/how-to-import-python-files/
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname("../"))))
    from controller import Controller
    from runtime.config import RuntimeConfig
    from runtime.opts import RuntimeOpts

    if os.environ.get("DEBUG"):
        import utils.debug as debug

DEMO_JSON_DATA = [
    {
        "abstract": "Python 3.9.1 - Python 3.8.7 - Mac OS X - Python 3.9.0 - Python 3.7.9",
        "sitelinks": [
            {
                "abstract": "Python 3.9.1 - Python 3.8.7 - Mac OS X - Python 3.9.0 - Python 3.7.9",
                "index": "1a",
                "title": "Downloads",
                "url": "https://www.python.org/downloads/",
            },
            {
                "abstract": "3.9.1 is the first version of Python to support macOS 11 Big Sur.",
                "index": "1b",
                "title": "Python 3.9.1",
                "url": "https://www.python.org/downloads/release/python-391/",
            },
            {
                "abstract": "Python's documentation, tutorials, and guides are constantly ...",
                "index": "1c",
                "title": "Documentation",
                "url": "https://www.python.org/doc/",
            },
            {
                "abstract": "Tutorial - Standard Library - Python Language - Python Module Index",
                "index": "1d",
                "title": "Python Docs",
                "url": "https://docs.python.org/",
            },
            {
                "abstract": "BeginnersGuide/Download - Python for Programmers - IDEs",
                "index": "1e",
                "title": "Python For Beginners",
                "url": "https://www.python.org/about/gettingstarted/",
            },
            {
                "abstract": "1. Whetting Your Appetite - 5. Data Structures - 9. Classes - ...",
                "index": "1f",
                "title": "Tutorial",
                "url": "https://docs.python.org/3/tutorial/",
            },
        ],
        "title": "Downloads",
        "url": "https://www.python.org/downloads/",
    },
    {
        "abstract": "Python is an interpreted, high-level and general-purpose programming language. Python's design philosophy emphasizes code readability with its notable use ...",
        "matches": [
            {"offset": 0, "phrase": "Python"},
            {"offset": 79, "phrase": "Python's"},
        ],
        "metadata": "Developer: Python Software Foundation | Designed by: Guido van Rossum | Filename extensions: .py,.pyi,.pyc,.pyd,.pyo ... | Typing discipline: Duck, dynamic, gradual (since ...",
        "title": "Python (programming language) - Wikipedia",
        "url": "https://en.wikipedia.org/wiki/Python_(programming_language)",
    },
    {
        "abstract": "プログラミング言語 Pythonの紹介・ダウンロードとインストール・Python入門講座など.",
        "matches": [
            {"offset": 10, "phrase": "Python"},
            {"offset": 34, "phrase": "Python"},
        ],
        "title": "プログラミング言語 Python - python.jp",
        "url": "https://www.python.jp/",
    },
    {
        "abstract": "Python · Great first language · Large programming community · Excellent online documentation · Endless libraries and packages · World-wide popularity · Powerful ...",
        "matches": [{"offset": 0, "phrase": "Python"}],
        "title": "Python Courses & Tutorials | Codecademy",
        "url": "https://www.codecademy.com/catalog/language/python",
    },
    {
        "abstract": "Well organized and easy to understand Web building tutorials with lots of examples of how to use HTML, CSS, JavaScript, SQL, PHP, Python, Bootstrap, Java ...",
        "matches": [{"offset": 130, "phrase": "Python"}],
        "title": "Python Tutorial - W3Schools",
        "url": "https://www.w3schools.com/python/",
    },
    {
        "abstract": "The Python Package Index (PyPI) is a repository of software for the Python programming language.",
        "matches": [
            {"offset": 4, "phrase": "Python"},
            {"offset": 68, "phrase": "Python"},
        ],
        "title": "PyPI · The Python Package Index",
        "url": "https://pypi.org/",
    },
]


def main(args=None):
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    warnings.simplefilter("ignore", FutureWarning)

    _ = RuntimeOpts.get_instance(args)
    runtime_config = RuntimeConfig.get_instance()
    if not runtime_config.valid():
        print("Config is invalid.")
        for error in runtime_config.errors:
            print("Error: %s" % error)
        sys.exit(1)

    data = DEMO_JSON_DATA
    error = None
    controller = Controller(data)
    try:
        _ = controller.run()
    except curses.error as e:
        error = e
    except Exception as e:
        error = e
    finally:
        controller._end_curses()
        if error != None:
            print("Error: %s" % error)
            sys.exit(1)


if __name__ == "__main__":
    main()



from fuzzywuzzy import process as fuzzyprocess
from itertools import groupby
import re
import os, sys

# local

try:
    # need when 「python3 gfzs/model.py」
    if __name__ == "__main__":
        # https://codechacha.com/ja/how-to-import-python-files/
        sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
        from utils import debug
        from runtime.opts import RuntimeOpts

    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    else:
        from gfzs.utils import debug
        from gfzs.runtime.opts import RuntimeOpts

# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    from utils import debug
    from runtime.opts import RuntimeOpts


class Model:
    # e.g.) collection = [{ title, url, abstract }, ...]
    def __init__(self, collection):
        self.collection = collection
        self.result = []
        self.query = self.old_query = None
        self.runtime_opts = RuntimeOpts.get_instance()
        self.errors = []

        self.char_regex = re.compile(r"^\w|\W+")
        # https://qiita.com/ganariya/items/42fc0ed3dcebecb6b117
        self.code_regex = re.compile(
            "[!\"#$%&'\\\\()*+,-./:;<=>?@[\\]^_`{|}~「」〔〕“”〈〉『』【】＆＊・（）＄＃＠。、？！｀＋￥％]"
        )
        self.space_regex = re.compile(r"^[ 　]")
        self._result_nested_by_url = self._collection_nested_by("url")

    @property
    def data_size(self):
        return len(self.result)

    def update_query(self, query):
        self.old_query = self.query
        self.query = query

    def push_query(self, char):
        self.old_query = self.query
        self.query += char

    def valid(self) -> bool:
        if not self._validate_json(self.collection):
            self.errors.append(
                Exception(
                    "Invalid json format. Please pass in an array of json that keeps `title`, `url` and `abstract` as keys."
                )
            )
            return False
        elif not self._validate_blank(self.collection):
            self.errors.append(Exception("The result passed was empty."))
            return False
        else:
            return True

    def validate_query(self, query=None) -> bool:
        def fn(q):
            if q is None or q is "\0":
                return False

            trimed = self.code_regex.sub("", q)
            is_not_empty = trimed != ""

            if is_not_empty:
                m = self.space_regex.match(trimed)
                is_space_start = m != None
                if is_space_start:
                    return False

                m = self.char_regex.match(trimed)
                is_char = m != None
                return is_char
            else:
                return False

        if query is None:
            return fn(self.query)
        else:
            return fn(query)

    def find(self, query=None):
        score = self.runtime_opts.score

        if query != None and query != "":
            self.update_query(query)

        if self.validate_query():
            self.result = self._make_result_from_scored(score)
        else:
            self.result = self.collection

        return self.result

    def _validate_json(self, data) -> bool:
        result = []

        for item in data:
            if "url" in item:
                result.append(True)
            else:
                result.append(False)

            if "abstract" in item:
                result.append(True)
            else:
                result.append(False)

            if "url" in item:
                result.append(True)
            else:
                result.append(False)

        return all(result)

    def _validate_blank(self, data) -> bool:
        if len(data) != 0:
            return True
        else:
            return False

    def _make_result_from_scored(self, score) -> list:
        result = []

        # From title
        scored_result_at_title = self._make_scored_result("title", score)
        # From abstract
        scored_result_at_abstract = self._make_scored_result("abstract", score)

        all_urls = list(set(scored_result_at_title.keys())) + list(
            set(scored_result_at_abstract.keys())
        )

        added = []
        for url in set(all_urls):
            score_at_title = 0
            score_at_abstract = 0
            title_len = 1
            abstract_len = 1

            if url in scored_result_at_title:
                data_at_title = scored_result_at_title[url]
                score_at_title = data_at_title["score"]
                title_len = len(data_at_title["data"]["title"])

            if url in scored_result_at_abstract:
                data_at_abstract = scored_result_at_abstract[url]
                score_at_abstract = data_at_abstract["score"]
                abstract_len = len(data_at_abstract["data"]["abstract"])

            score = (title_len / (title_len + abstract_len)) * score_at_title + (
                abstract_len / (title_len + abstract_len)
            ) * score_at_abstract

            added.append({"score": score, "url": url})

        added = sorted(added, key=lambda x: x["score"], reverse=True)
        for item in added:
            url = item["url"]
            result.append(self._result_nested_by_url[url])

        return result

    def _make_scored_result(self, key, score) -> dict:
        result = {}

        # From key
        data_grouped_by_key = self._collection_nested_by(key)
        keys = data_grouped_by_key.keys()
        fuzzysorted = fuzzyprocess.extract(self.query, keys, limit=len(keys))

        # fuzzysorted reutrns an array of tuples: [('one', 45), ('three', 45), ('two', 0)]
        for item in fuzzysorted:
            if item[1] > score:
                scored = {}

                match_data = data_grouped_by_key.get(item[0])
                url = match_data["url"]
                result[url] = {"data": match_data, "score": item[1]}

        return result

    def _collection_nested_by(self, key) -> dict:
        data = {}

        for c in self.collection:
            key_data = c[key]
            data[key_data] = c

        return data


if __name__ == "__main__":

    class TestModel(Model):
        @property
        def summary_count(self):
            return len(self.collection)


if __name__ == "__main__":
    import json
    import signal
    import argparse

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    from runtime.opts import RuntimeOpts

    json_str = open("fixtures/rust.json", "r").read()
    data = json.loads(json_str)

    args = argparse.Namespace(score=30)
    config = RuntimeOpts.get_instance(args)
    model = TestModel(data)

    result = model.find("Amazon")
    print(
        "Search (query=Amazon, score=%d):  %d / %d"
        % (config.score, model.data_size, model.summary_count)
    )
    for i in range(len(result)):
        print(result[i]["title"])

    result = model.find("\0")
    print(
        "Search (query=\0, score=%d):  %d / %d"
        % (config.score, model.data_size, model.summary_count)
    )
    for i in range(len(result)):
        print(result[i]["title"])

    result = model.find("a")
    print(
        "Search (query=\0, score=%d):  %d / %d"
        % (config.score, model.data_size, model.summary_count)
    )
    for i in range(len(result)):
        print(result[i]["title"])



import curses
import os, sys

GOOGLE = "Google"
FUZZY = "Fuzzy"
SEARCH = "Search"

try:
    # need when 「python3 gfzs/views/header.py」
    if __name__ == "__main__":
        # https://codechacha.com/ja/how-to-import-python-files/
        sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
        import info
        from utils import debug

        from base import Base

        if os.environ.get("DEBUG"):
            import utils.debug as debug
    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    else:
        from gfzs import info
        from gfzs.utils import debug
        from gfzs.views.base import Base

        if os.environ.get("DEBUG"):
            import gfzs.utils.debug as debug

# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    # https://codechacha.com/ja/how-to-import-python-files/
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname("../"))))
    import info
    from utils import debug
    from views.base import Base

    if os.environ.get("DEBUG"):
        import utils.debug as debug


class Header(Base):
    def __init__(self, stdscr):
        super().__init__(stdscr, None, "header")
        self.version = "(%s)" % info.__version__
        self.copyright = info.__copyright__

    def create(self):
        self._init_layout()
        self._make_header()
        self.window.refresh()

    def reset(self):
        self.window.erase()
        self._init_layout()
        self._make_header()
        self.window.refresh()

    def _init_layout(self):
        self.parent_height, self.parent_width = self.stdscr.getmaxyx()
        self.window = curses.newwin(2, self.parent_width, 0, 0)

    # https://stackoverflow.com/a/53016371/9434894
    def _make_header(self):
        start_index = 0
        # Write Google
        google = list(GOOGLE)
        first_o = True
        for i in range(len(google)):
            c = google[i]
            if c == "o":
                if first_o:
                    first_o = False
                    self.window.addstr(0, i, c, self.color.google("o"))
                else:
                    self.window.addstr(0, i, c, self.color.google("o2"))
            else:
                self.window.addstr(0, i, c, self.color.google(c))

        # Write Fuzzy
        start_index += len(GOOGLE) + 1
        self.window.addstr(0, start_index, FUZZY, self.color.fuzzy())

        # Write Search
        start_index += len(FUZZY) + 1
        self.window.addstr(0, start_index, SEARCH, self.color.search())

        # Write verion
        start_index += len(SEARCH) + 1
        self.window.addstr(0, start_index, self.version, self.color.version())

        # Write Copyright
        self.window.addstr(
            0,
            self.parent_width - len(self.copyright),
            self.copyright,
            self.color.copy_right(),
        )
        self.window.hline(
            1, 0, curses.ACS_HLINE | self.colors["hline"], self.parent_width
        )


if __name__ == "__main__":

    class TestHeader(Header):
        def run(self):
            self._loop()

        def _end_curses(self):
            """ Terminates the curses application. """
            curses.nocbreak()
            self.window.keypad(0)
            curses.echo()
            curses.endwin()

        def _loop(self):
            self.create()

            while True:
                try:
                    user_input = self.window.getch()
                except curses.error:
                    continue
                except KeyboardInterrupt:
                    break

                if user_input == curses.KEY_RESIZE:
                    self.reset()


if __name__ == "__main__":
    import curses
    import signal
    import os, sys

    # local

    # https://codechacha.com/ja/how-to-import-python-files/
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
    import utils, model
    from model import Model
    from runtime.config import RuntimeConfig

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    runtime_config = RuntimeConfig.get_instance()
    if not runtime_config.valid():
        print("Config is invalid.")
        for error in runtime_config.errors:
            print("Error: %s" % error)
        sys.exit(1)

    # initscr() returns a window object representing the entire screen.
    stdscr = curses.initscr()

    # turn off automatic echoing of keys to the screen
    curses.noecho()
    # Buffering off
    # https://docs.python.org/ja/3/library/curses.html#curses.cbreak
    curses.cbreak()
    # Disable the mouse cursor.
    curses.curs_set(0)

    target = TestHeader(stdscr)
    error = None
    try:
        target.run()
    except curses.error as e:
        error = str(e)
    finally:
        target._end_curses()
        if error != None:
            print(error)



# copy from: https://qiita.com/t4t5u0/items/bd5270d0a7b3def35a03

import unicodedata
import os, sys
import curses

# from local

try:
    # need when「python3 gfzs/multibyte.py」
    if __name__ == "__main__":
        # https://codechacha.com/ja/how-to-import-python-files/
        sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
        import debug
    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    else:
        from gfzs.utils import debug
# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    # https://codechacha.com/ja/how-to-import-python-files/
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname("../"))))
    import utils
    from utils import debug


# https://seiai.ed.jp/sys/text/cs/mcodes/ucodeutf8.html
INVALID_BYTE_START = 0x80
INVALID_BYTE_END = 0xBF
BYTE1_START = 0x00
BYTE1_END = 0x7F
BYTE2_START = 0xC0
BYTE2_END = 0xDF
BYTE3_START = 0xE0
BYTE3_END = 0xEF
BYTE4_START = 0xF0
BYTE4_END = 0xFF

# ref: https://qiita.com/masakielastic/items/8eb4bf4efc2310ee7baf


class Multibyte:
    def __init__(self, stdscr=None):
        self.stdscr = stdscr

    # https://note.nkmk.me/python-unicodedata-east-asian-width-count/
    def get_east_asian_width_count(self, text):
        count = 0
        for c in text:
            if self.is_full_width(c):
                count += 2
            else:
                count += 1
        return count

    def is_full_width(self, c):
        return unicodedata.east_asian_width(c) in "FWA"

    def marked_full_width(self, text, mark="\0") -> str:
        result = ""
        for c in text:
            if self.is_full_width(c):
                result += "%s%s" % (c, mark)
            else:
                result += c

        return result

    def unmarked_full_width(self, text, mark="\0") -> str:
        return text.replace(mark, "")

    def getch(self):
        key = self.stdscr.getch()
        text_pool = [key]
        codepoint = key

        if BYTE1_START <= key <= BYTE1_END:
            # e.g.) ascii
            pass
        elif INVALID_BYTE_START <= key <= INVALID_BYTE_END:
            self._end_curses()
            e = Exception("[Multibyte] Invalid Byte: %s" % key)
            print("Error: %s" % str(e))
            sys.exit(1)
        elif BYTE2_START <= key <= BYTE2_END:
            # e.g.) Umlaut
            text_pool.append(self.stdscr.getch())
            a, b = text_pool
            tmp = map(lambda x: bin(x)[2:], [0b00011111 & a, 0b00111111 & b])
            tmp = "".join(item.zfill(6) for item in tmp)
            codepoint = int(tmp, 2)
        elif BYTE3_START <= key <= BYTE3_END:
            # e.g.) Japanease
            for _ in range(2):
                text_pool.append(self.stdscr.getch())

            # e.g) 'あ' = [227, 129, 130]
            a, b, c = text_pool
            tmp = map(
                lambda x: bin(x)[2:], [0b00001111 & a, 0b00111111 & b, 0b00111111 & c]
            )
            tmp = "".join([item.zfill(6) for item in tmp])
            codepoint = int(tmp, 2)
        elif BYTE4_START <= key <= BYTE4_END:
            for _ in range(3):
                text_pool.append(self.stdscr.getch())

            a, b, c, d = text_pool
            tmp = map(
                lambda x: bin(x)[2:],
                [0b00000111 & a, 0b00111111 & b, 0b00111111 & c, 0b00111111 & d],
            )
            tmp = "".join([item.zfill(6) for item in tmp])
            codepoint = int(tmp, 2)
        else:
            pass

        return codepoint

    def _end_curses(self):
        """ Terminates the curses application. """
        curses.nocbreak()
        curses.echo()
        curses.endwin()



import curses
import curses.ascii
import webbrowser
import os

# local

try:
    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    from gfzs.model import Model
    from gfzs.utils.multibyte import Multibyte
    from gfzs.views.header import Header
    from gfzs.views.search_result import SearchResult
    from gfzs.views.footer import Footer

    if os.environ.get("DEBUG"):
        import gfzs.utils.debug as debug

# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    import utils, views
    from model import Model
    from utils.multibyte import Multibyte
    from views.header import Header
    from views.search_result import SearchResult
    from views.footer import Footer

    if os.environ.get("DEBUG"):
        from utils import debug

KEY_ENTER = 10
KEY_ESC = 27


class Controller:
    def __init__(self, data):
        self._init_curses()
        self.model = Model(data)
        self.header = Header(self.stdscr)
        self.search_result = SearchResult(self.stdscr, self.model)
        self.footer = Footer(self.stdscr, self.model)
        self.multibyte = Multibyte(self.stdscr)

    def _init_curses(self):
        """ Inits the curses application """
        # initscr() returns a window object representing the entire screen.
        self.stdscr = curses.initscr()
        # turn off automatic echoing of keys to the screen
        curses.noecho()
        # Enable non-blocking mode. keys are read directly, without hitting enter.
        curses.cbreak()
        # Able mouse cursor
        curses.curs_set(1)
        self.stdscr.keypad(1)
        # Enable colorous output.
        self.stdscr.refresh()

    def _end_curses(self):
        """ Terminates the curses application. """
        curses.nocbreak()
        self.stdscr.keypad(0)
        curses.echo()
        curses.endwin()

    def _search_and_refresh_display(
        self, user_input, is_init_property=False, is_init_query=True
    ):
        # Disable mouse cursor
        curses.curs_set(0)

        self.search_result.reset()

        if is_init_property:
            self.search_result.init_properties_after_create()

        result = self.search_result.update_view_in_loop()

        if is_init_query:
            self.search_result.update_query("")

        if result:
            self.search_result.handle_key_in_loop(user_input)

    def _handle_resize(self, user_input):
        self.header.reset()
        self.footer.reset()
        self._search_and_refresh_display(
            user_input, is_init_property=True, is_init_query=False
        )

    def execute_when_enter(self, current_selected):
        result = self.model.result
        if len(result) != 0:
            webbrowser.open(result[current_selected].get("url"), new=2)

    def run(self) -> int:
        input_mode = True
        user_input = ""
        box_select_begin_y = 2
        arrow_keys = (curses.KEY_DOWN, curses.KEY_UP, curses.KEY_LEFT, curses.KEY_RIGHT)
        backspace_keys = (curses.ascii.BS, curses.ascii.DEL, curses.KEY_BACKSPACE)

        self.header.create()

        self.search_result.create(box_select_begin_y)
        self.search_result.init_properties_after_create()
        self.search_result.update_view_in_loop()

        self.footer.create()
        self.footer.activate(is_init=True)

        while True:
            # stdscr.refresh is called in the process of updating the query of footer and disappears at that time
            self.header.create()

            if input_mode:
                self.footer.activate()
            else:
                self.search_result.update_view_in_loop()

            try:
                user_input = self.multibyte.getch()
            except curses.error:
                continue
            except KeyboardInterrupt:
                self._end_curses()
                break

            if input_mode:
                if user_input in arrow_keys:
                    input_mode = False
                    self._search_and_refresh_display(user_input, is_init_query=False)
                elif user_input == KEY_ESC:
                    pass
                elif user_input == KEY_ENTER:
                    input_mode = False
                    self._search_and_refresh_display(user_input, is_init_query=False)
                # https://www.programcreek.com/python/?code=mingrammer%2Fawesome-finder%2Fawesome-finder-master%2Fawesome%2Ftui.py
                elif user_input in backspace_keys:
                    if self.model.query == "":
                        pass
                    else:
                        self.footer.delete_char()
                        self._search_and_refresh_display(
                            user_input, is_init_property=True, is_init_query=False
                        )
                elif user_input == curses.KEY_RESIZE:
                    self._handle_resize(user_input)
                # I don't know the reason, but - 1 may come in
                elif user_input == -1:
                    pass
                else:
                    text = chr(user_input)
                    will_query = self.model.query + text
                    if self.model.validate_query(will_query):
                        self.footer.write(text)
                        self._search_and_refresh_display(
                            user_input, is_init_property=True, is_init_query=False
                        )
            else:
                if user_input in arrow_keys:
                    self.search_result.handle_key_in_loop(user_input)
                elif user_input == KEY_ESC:
                    pass
                elif user_input == KEY_ENTER:
                    self.execute_when_enter(self.search_result.current_selected)
                elif user_input in backspace_keys:
                    input_mode = True
                    self.footer.activate()
                    self.footer.delete_char()
                    self._search_and_refresh_display(
                        user_input, is_init_property=True, is_init_query=False
                    )
                elif user_input == curses.KEY_RESIZE:
                    self._handle_resize(user_input)
                # I don't know the reason, but - 1 may come in
                elif user_input == -1:
                    pass
                else:
                    input_mode = True
                    text = chr(user_input)
                    self.footer.activate()
                    self.footer.write(text)


if __name__ == "__main__":
    import json
    import os, sys
    import signal

    # https://note.nkmk.me/python-warnings-ignore-warning/
    import warnings

    # local

    # https://codechacha.com/ja/how-to-import-python-files/
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
    from runtime.config import RuntimeConfig

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    warnings.simplefilter("ignore", FutureWarning)

    runtime_config = RuntimeConfig.get_instance()
    if not runtime_config.valid():
        print("Config is invalid.")
        for error in runtime_config.errors:
            print("Error: %s" % error)
        sys.exit(1)

    json_str = open("fixtures/rust.json", "r").read()
    data = json.loads(json_str)

    error = None
    controller = Controller(data)
    try:
        choice = controller.run()
        result = controller.model.result
        if not choice is None:
            print(result[choice].get("title"))
    except curses.error as e:
        error = str(e)
    finally:
        controller._end_curses()
        if error != None:
            print(error)



# ref: https://github.com/NikolaiT/Scripts/blob/master/scripts/python/curses/text_selector.py

import curses
import unicodedata
import textwrap
import math
import os, sys

# local

try:
    # need when 「python3 gfzs/views/search_result.py」
    if __name__ == "__main__":
        # https://codechacha.com/ja/how-to-import-python-files/
        sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
        from utils.markup import Markup
        from utils.multibyte import Multibyte

        from not_found import NotFound
        from paging import Paging
        from base import Base

        if os.environ.get("DEBUG"):
            import utils.debug as debug
    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    else:
        from gfzs.utils.markup import Markup
        from gfzs.utils.multibyte import Multibyte

        from gfzs.views.not_found import NotFound
        from gfzs.views.paging import Paging
        from gfzs.views.base import Base

        if os.environ.get("DEBUG"):
            import gfzs.utils.debug as debug
# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname("../"))))
    from utils.markup import Markup
    from utils.multibyte import Multibyte

    from views.not_found import NotFound
    from views.paging import Paging
    from views.base import Base

    if os.environ.get("DEBUG"):
        import utils.debug as debug


class SearchResultHelper:
    def __init__(self):
        self.current_selected = 0
        self.last = 1
        self.topy = 1
        self.top_textbox = None
        self.pad_begin_y = 0
        self.per_page = 1
        self.change_page = False

    def update_attributues(self, current_selected, last, topy, top_textbox):
        self.current_selected = current_selected
        self.last = last
        self.topy = topy
        self.top_textbox = top_textbox

    def update_per_page(self, value):
        self.per_page = value

    @property
    def current_page(self) -> int:
        return self.current_selected // self.per_page + 1

    def prev_page(self, total_data_size):
        per_page = self.per_page
        old_current_selected = self.current_selected
        current_pagetop_index = per_page * math.floor(old_current_selected / per_page)

        self.change_page = True
        if self.current_page == 1:
            total_page = math.ceil(total_data_size / per_page)

            if total_page >= 1:
                self.current_selected = (total_page - 1) * per_page
            else:
                pass
        else:
            self.current_selected = current_pagetop_index - per_page

    def next_page(self, total_data_size):
        per_page = self.per_page
        old_current_selected = self.current_selected
        current_pagetop_index = per_page * math.floor(old_current_selected / per_page)
        total_page = math.ceil(total_data_size / per_page)

        self.change_page = True
        if self.current_page == total_page:
            self.current_selected = 0
        else:
            self.current_selected = current_pagetop_index + per_page

    def down(self, total_data_size):
        old_current_page = self.current_page
        old_current_selected = self.current_selected

        if old_current_selected >= total_data_size - 1:
            self.change_page = True
            self.current_selected = 0
        else:
            self.current_selected += 1
            if self.current_page != old_current_page:
                self.change_page = True

    def up(self, total_data_size):
        old_current_page = self.current_page
        old_current_selected = self.current_selected

        if old_current_selected == 0:
            self.change_page = True
            self.current_selected = total_data_size - 1  # wrap around.
        else:
            self.current_selected -= 1
            if self.current_page != old_current_page:
                self.change_page = True


class SearchResult(Base):
    """Display options build from a list of strings in a (unix) terminal.
    The user can browser though the textboxes and select one with enter.
    """

    def __init__(self, stdscr, model):
        """Create a SearchResult object.
        'data' is list of string. Each string is used to build
        a textbox.
        """
        super().__init__(stdscr, model, "search_result")
        self.paging = Paging(stdscr, self)
        self.stop_loop = False
        self.textboxes = []
        self.helper = SearchResultHelper()
        self.not_found = NotFound(stdscr)
        self.markup = Markup()
        self.multibyte = Multibyte()

        # Element parameters. Channge them here.
        self.TEXTBOX_HEIGHT = 8
        self.PAD_WIDTH = 400
        self.PAD_HEIGHT = 1000

    @property
    def current_selected(self) -> int:
        return self.helper.current_selected

    @property
    def per_page(self) -> int:
        return self.helper.per_page

    @property
    def data_size(self) -> int:
        return self.model.data_size

    @property
    def current_page(self) -> int:
        return self.helper.current_page

    def update_per_page(self, value):
        self.helper.update_per_page(value)

    def init_properties_after_create(self):
        current_selected = 0
        textboxes_len = len(self.textboxes)

        if textboxes_len > 0:
            top_textbox = self.textboxes[0]
            topy, _ = self._refresh_view(top_textbox)
            if textboxes_len > 1:
                last = 1
            else:
                last = 0

            self.helper.update_attributues(current_selected, last, topy, top_textbox)

    def create(self, pad_begin_y=0):
        self.helper.pad_begin_y = pad_begin_y
        self._init_layout()
        self._create_pad()
        self.textboxes = self._make_textboxes()
        if len(self.textboxes) > 0:
            self._refresh_view(self.textboxes[0])
            self.paging.create()

    def destroy(self):
        self._delete_pad()

    def update_query(self, query):
        self.model.update_query(query)

    def reset(self):
        self._reset_pad()
        self._init_layout()
        self.textboxes = self._make_textboxes()
        if len(self.textboxes) > 0:
            self.not_found.destroy()
            self._refresh_view(self.textboxes[0])
            self.paging.create()
        else:
            self.paging.destroy()
            self.destroy()
            self.not_found.create()

    def _init_layout(self):
        self.parent_height, self.parent_width = self.stdscr.getmaxyx()
        # 「2」 is the height of the header or footer
        # 「4」 = header + footer height
        self.window = curses.newwin(self.parent_height - 4, self.parent_width, 2, 0)
        self.update_per_page(self.parent_height // self.TEXTBOX_HEIGHT - 1)
        # https://stackoverflow.com/a/17369532/9434894
        self.window.keypad(1)

    def _delete_pad(self):
        self.pad.clear()
        self.window.refresh()

    def _reset_pad(self):
        self._delete_pad()
        self._init_layout()
        self._create_pad()

    def _create_pad(self):
        """ Creates a big self.pad to place the textboxes in. """
        self.pad = curses.newpad(self.PAD_HEIGHT, self.PAD_WIDTH)
        self.pad.box()
        self.window.refresh()

    def _make_textboxes(self):
        """Build the textboxes in the pad center and put them in the
        horizontal middle of the pad."""
        textboxes = []
        i = 1
        data = self.model.find()

        if len(data) == 0:
            return textboxes

        for s in data:
            textbox = self.pad.derwin(
                self.TEXTBOX_HEIGHT,
                self.parent_width - 4,
                i + self.helper.pad_begin_y,
                2,
            )

            textboxes.append(textbox)
            i += self.TEXTBOX_HEIGHT

        # When all are displayed as multi-byte character strings
        gap = 6  # 4 = 1 (Frame border) + 3(padding) + 2(margin)
        abstract_line_len = self.parent_width // 2 - gap
        for k in range(len(textboxes)):
            textboxes[k].box()

            title = data[k].get("title")
            url = data[k].get("url")
            abstract = data[k].get("abstract")

            textboxes[k].addstr(
                2, 2, "%s%-3s" % ("", str(k + 1) + "."), self.colors["index"]
            )
            textboxes[k].addstr(2, 6, title, self.colors["title"])
            textboxes[k].addstr(3, 6, url, self.colors["url"])
            lines = textwrap.wrap(abstract, abstract_line_len)
            for l in range(len(lines)):
                textboxes[k].addstr(4 + l, 6, lines[l], self.colors["abstract"])

            # Markup Search Query for title
            markup_data = self.markup.parse(title, self.model.query)
            for search_text in markup_data:
                for item in markup_data[search_text]:
                    offset_x = item["half_width"]["start_index"]
                    color = item["color"]
                    match_text = item["match"]
                    textboxes[k].addstr(
                        2, 6 + offset_x, match_text, color | curses.A_BOLD
                    )

            # Markup Search Query for abstract
            for l in range(len(lines)):
                markup_data = self.markup.parse(lines[l], self.model.query)
                for search_text in markup_data:
                    for item in markup_data[search_text]:
                        offset_x = item["half_width"]["start_index"]
                        color = item["color"]
                        match_text = item["match"]

                        match_text_byte_len = self.multibyte.get_east_asian_width_count(
                            match_text
                        )
                        if offset_x + 6 + match_text_byte_len <= self.parent_width - 1:
                            textboxes[k].addstr(
                                4 + l, 6 + offset_x, match_text, color | curses.A_BOLD
                            )
                        # Wrap display
                        else:
                            match_text_before = match_text[
                                : (self.parent_width - offset_x)
                            ]
                            byte_len = self.multibyte.get_east_asian_width_count(
                                match_text_before
                            )
                            gap = 6  # 4 = 1 (Frame border) + 3(padding) + 2(margin)
                            textboxes[k].addstr(
                                4 + l,
                                self.parent_width - byte_len - gap,
                                match_text_before,
                                color,
                            )

                            match_text_after = match_text[
                                (self.parent_width - offset_x) :
                            ]
                            textboxes[k].addstr(5 + l, 6, match_text_after, color)

        return textboxes

    def _refresh_view(self, window):
        """ Refresh textboxes """
        cy, cx = window.getbegyx()

        per_page = self.per_page
        display_limit_pos_y = self.TEXTBOX_HEIGHT * per_page
        display_limit_pos_x = self.parent_width - 1

        # Since the display of the last pad is cut off, display_limit_pos_y + self.helper.pad_begin_y
        display_limit_pos_y += self.helper.pad_begin_y

        self.pad.refresh(
            cy,
            cx,
            1 + self.helper.pad_begin_y,
            2,
            display_limit_pos_y,
            display_limit_pos_x,
        )
        return (cy, cx)

    def update_view_in_loop(self) -> bool:
        textboxes = self.textboxes
        textboxes_len = len(textboxes)

        if textboxes_len == 0:
            return False

        current_selected = self.current_selected
        last = self.helper.last
        topy = self.helper.topy
        top_textbox = self.helper.top_textbox

        # Highligth the selected one, the last selected textbox should
        # become normal again.
        if last == current_selected == 0:
            textboxes[current_selected].border(self.color.highlight)
        else:
            textboxes[current_selected].border(self.color.highlight)
            textboxes[last].border()

        # Paging

        # While the textbox can be displayed on the page with the current top_textbox,
        # don't after the view. When this becomes impossible,
        # center the view to last displayable textbox on the previous view.
        cy, cx = textboxes[current_selected].getbegyx()
        per_page = self.per_page

        # The current window is to far down. Switch the top textbox.
        # When you reach the bottom, redisplay the current box at the top
        if (topy + self.parent_height - self.TEXTBOX_HEIGHT) <= cy:
            top_textbox = textboxes[current_selected]

        # The current window is to far up. There is a better way though...
        # Update the top until you reach the top of the screen.
        if topy >= cy + self.TEXTBOX_HEIGHT:
            top_textbox = textboxes[current_selected]

        if last == 0:
            if current_selected > per_page + 1:
                top_textbox = textboxes[current_selected - per_page + 1]

        if self.helper.change_page:
            top_textbox = textboxes[current_selected]

        if last != current_selected:
            last = current_selected

        refresh_topy, _ = self._refresh_view(top_textbox)

        if self.helper.change_page:
            self.paging.reset()

        self.textboxes = textboxes

        self.update_per_page(per_page)
        self.helper.update_attributues(
            current_selected, last, refresh_topy, top_textbox
        )

        return True

    def handle_key_in_loop(self, user_input):
        self.helper.change_page = False
        textboxes = self.textboxes
        textboxes_len = len(textboxes)

        if textboxes_len == 0:
            return

        per_page = self.per_page

        if textboxes_len > 1 and user_input == curses.KEY_DOWN:
            self.helper.down(textboxes_len)
        elif textboxes_len > 1 and user_input == curses.KEY_UP:
            self.helper.up(textboxes_len)
        elif textboxes_len > per_page + 1 and user_input == curses.KEY_RIGHT:
            self.helper.next_page(textboxes_len)
        elif textboxes_len > per_page + 1 and user_input == curses.KEY_LEFT:
            self.helper.prev_page(textboxes_len)
        elif user_input == curses.KEY_RESIZE:
            self.reset()
        elif user_input in (curses.ascii.BS, curses.ascii.DEL, curses.KEY_BACKSPACE):
            self.reset()
        elif user_input == ord("q"):  # Quit without selecting.
            self.stop_loop = True


if __name__ == "__main__":

    class TestSearchResult(SearchResult):
        def run(self, pad_begin_y=0):
            """ Just run this when you want to spawn the selection process. """
            self.create(pad_begin_y)
            picked = self._loop()
            self._end_curses()
            return picked

        def _end_curses(self):
            """ Terminates the curses application. """
            curses.nocbreak()
            self.window.keypad(0)
            curses.echo()
            curses.endwin()

        def _loop(self):
            current_selected = 0
            textboxes_len = len(self.textboxes)

            if textboxes_len > 1:
                last = 1
            else:
                last = 0

            # See at the root textbox.
            if textboxes_len > 0:
                top_textbox = self.textboxes[0]
                topy, _ = self._refresh_view(top_textbox)
                self.helper.update_attributues(
                    current_selected, last, topy, top_textbox
                )

            while True:
                if self.stop_loop:
                    break

                self.update_view_in_loop()

                try:
                    user_input = self.window.getch()
                except curses.error:
                    continue
                except KeyboardInterrupt:
                    break

                if user_input == curses.KEY_ENTER or user_input == 10:
                    return int(self.current_selected)

                self.handle_key_in_loop(user_input)


if __name__ == "__main__":
    import curses
    import signal
    import os, sys
    import json

    # local

    # https://codechacha.com/ja/how-to-import-python-files/
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
    from model import Model
    from runtime.config import RuntimeConfig

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    runtime_config = RuntimeConfig.get_instance()
    if not runtime_config.valid():
        print("Config is invalid.")
        for error in runtime_config.errors:
            print("Error: %s" % error)
        sys.exit(1)

    json_str = open("fixtures/rust.json", "r").read()
    data = json.loads(json_str)

    # initscr() returns a window object representing the entire screen.
    stdscr = curses.initscr()

    # turn off automatic echoing of keys to the screen
    curses.noecho()
    # Enable non-blocking mode. keys are read directly, without hitting enter.
    curses.cbreak()
    # Disable the mouse cursor.
    curses.curs_set(0)

    stdscr.refresh()

    model = Model(data)
    model.update_query("Amazon")
    _ = model.find()

    target = TestSearchResult(stdscr, model)
    error = None
    try:
        choice = target.run(pad_begin_y=1)
        if choice != None:
            print(model.result[choice].get("title"))
    except curses.error as e:
        error = str(e)
    finally:
        if error != None:
            print(error)



# ref: https://qiita.com/ttsubo/items/c4af71ceba15b5b213f8
import argparse
import curses
import os
import sys
import json
from flatten_dict import flatten

# local

try:
    # need when 「python3 gfzs/runtime/config.py」
    if __name__ == "__main__":
        # https://codechacha.com/ja/how-to-import-python-files/
        sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
        from utils import debug

    # need when 「cat fixtures/rust.json | python -m gfzs」
    # need when 「cat fixtures/rust.json | bin/gfzs」
    else:
        from gfzs.utils import debug

# need when 「python3 gfzs/controller.py」
except ModuleNotFoundError:
    sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname("../"))))
    from utils import debug

# ~/.gfzsrc
DEFAULT_CONFIG_PATH = "%s/.gfzsrc" % os.path.expanduser("~")

# 0: curses.COLOR_BLACK
# 1: curses.COLOR_RED
# 2: curses.COLOR_GREEN
# 3: curses.COLOR_YELLOW
# 4: curses.COLOR_BLUE
# 5: curses.COLOR_MAGENTA
# 6: curses.COLOR_CYAN
# 7: curses.COLOR_WHITE
DEFAULT_CONFIG = {
    "view": {
        "footer": {
            "message": "QUERY>",
            "color": {
                "message": {
                    "text": curses.COLOR_GREEN,
                    "background": curses.COLOR_BLACK,
                    "style": "normal",
                },
                "hline": {
                    "text": curses.COLOR_WHITE,
                    "background": curses.COLOR_BLACK,
                    "style": "normal",
                },
            },
        },
        "header": {
            "color": {
                "hline": {
                    "text": curses.COLOR_WHITE,
                    "background": curses.COLOR_BLACK,
                    "style": "normal",
                },
            }
        },
        "search_result": {
            "color": {
                "index": {
                    "text": curses.COLOR_CYAN,
                    "background": curses.COLOR_BLACK,
                    "style": "normal",
                },
                "title": {
                    "text": curses.COLOR_GREEN,
                    "background": curses.COLOR_BLACK,
                    "style": "bold",
                },
                "url": {
                    "text": curses.COLOR_YELLOW,
                    "background": curses.COLOR_BLACK,
                    "style": "link",
                },
                "abstract": {
                    "text": curses.COLOR_WHITE,
                    "background": curses.COLOR_BLACK,
                    "style": "normal",
                },
                "markup_partial": {
                    "text": curses.COLOR_GREEN,
                    "background": curses.COLOR_MAGENTA,
                    "style": "normal",
                },
                "markup_char": {
                    "text": curses.COLOR_RED,
                    "background": curses.COLOR_BLACK,
                    "style": "normal",
                },
            }
        },
        "paging": {
            "color": {
                "common": {
                    "text": curses.COLOR_GREEN,
                    "background": curses.COLOR_BLACK,
                    "style": "bold",
                }
            }
        },
    }
}


class Singleton(object):
    @classmethod
    def get_instance(cls):
        if not hasattr(cls, "_instance"):
            cls._instance = cls()

        return cls._instance


class RuntimeConfig(Singleton):
    """A class that reads and manages the options passed in the runtime"""

    SUPPORT_FLATTEN_KEYS = list(flatten(DEFAULT_CONFIG, reducer="dot"))
    SUPPORT_COLOR_NUMBERS = list(range(8))
    SUPPORT_STYLES = ["normal", "link", "bold"]

    def __init__(self):
        self.data = self._create_data()
        self.errors = []

    @property
    def config_path(self):
        return DEFAULT_CONFIG_PATH

    def valid(self) -> bool:
        flatten_data = flatten(
            self.data, reducer="dot", keep_empty_types=(dict, str, int, list)
        )
        for flatten_key in flatten_data:
            target_val = flatten_data[flatten_key]
            # Check key
            if not flatten_key in RuntimeConfig.SUPPORT_FLATTEN_KEYS:
                self.errors.append(
                    Exception(
                        "Contains unsupported key.   (key_path, value) = (%s, %s)."
                        % (flatten_key, target_val)
                    )
                )
            # Check value
            else:
                if flatten_key.endswith("text") or flatten_key.endswith("background"):
                    if not target_val in RuntimeConfig.SUPPORT_COLOR_NUMBERS:
                        self.errors.append(
                            Exception(
                                "Contains unsupported value. (key_path, value) = (%s, %s)."
                                % (flatten_key, target_val)
                            )
                        )
                elif flatten_key.endswith("style"):
                    if not target_val in RuntimeConfig.SUPPORT_STYLES:
                        self.errors.append(
                            Exception(
                                "Contains unsupported value. (key_path, value) = (%s, %s)."
                                % (flatten_key, target_val)
                            )
                        )

        if self.errors != []:
            return False
        else:
            return True

    def _create_data(self) -> dict:
        """Load the app settings. If the config file does not exist, it will load the default config."""
        config_path = self.config_path

        data = dict
        if os.path.exists(config_path):
            try:
                json_str = open(config_path).read()
                data = json.loads(json_str)
            except json.decoder.JSONDecodeError as e:
                self._end_curses()
                raise e
        else:
            data = DEFAULT_CONFIG

        return data

    def _end_curses(self):
        """ Terminates the curses application. """
        curses.nocbreak()
        curses.echo()
        curses.endwin()


if __name__ == "__main__":

    config = RuntimeConfig.get_instance()
    print(config.data)
